#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <crypt.h>
#include <capi.h>
#include <keymgmt/keymgmt.h>
#include "defs.h"

/* The names of the test external public and private key files */

#define PGP_PUBKEY_FILE		"pubring.pgp"
#define PGP_PRIVKEY_FILE	"secring.pgp"
#define MAX_MPI_SIZE		512
#define MAX_USERID_SIZE		256

typedef struct {
	/* Key components (in PGP format) */
	BYTE n[ MAX_MPI_SIZE ], e[ MAX_MPI_SIZE ], d[ MAX_MPI_SIZE ];
	BYTE p[ MAX_MPI_SIZE ], q[ MAX_MPI_SIZE ], u[ MAX_MPI_SIZE ];
	int nLen, eLen, dLen, pLen, qLen, uLen;

	/* Key components (in cryptlib format) */
	CRYPT_PKCINFO_RSA rsaKey;

	/* userID for this key */
	char userID[ MAX_USERID_SIZE ];
	} PGP_INFO;

int getlen(i)
{
	if (i%8) return i/8+1;
	else return i/8;
}	

int lead_zero(i)
{	
	int nibbles = 0;
	
	if(i%4) nibbles = i/4+1;
	else nibbles = i/4;
	if (2*getlen(i) == nibbles) return 0;
	else return (1);
}

void getstr(char *str, mpz_t mp, int len)
{
	if (lead_zero(len)) 
	{
		*str = '0';
		mpz_get_str(str+1, 16, mp);
	}
	else	mpz_get_str(str, 16, mp);
}	

void getPassword( const char *prompt, char *password )
	{
	printf( "Please enter %s: ", prompt );
	fflush( stdout );
	fgets( password, 200, stdin );
	password[ strlen( password ) - 1 ] = '\0';
	}

void RSA_private_key_sanity(	mpz_t n,
				mpz_t e, 
				mpz_t p,
				mpz_t q, 
				mpz_t d, 
				mpz_t u, 
				mpz_t exp1, 
				mpz_t exp2, 
				mpz_t coeff);

int main( int argc, char **argv )
	{
	FILE *inFile, *pub, *sec, *out;
	char *encryptID = NULL; 
	char password[ 200 ], pubkey[200], seckey[200];
	int status, i, j;
	PGP_INFO	pgp_info;
	GETKEY_INFO	gkinfo;
		
	mpz_t	n, e, p, q, d, u, t, exp1, exp2, coeff;
	char	exp1_str[4096],
		exp2_str[4096],
		p_str[4096],
		q_str[4096],
		c_str[4096],
		d_str[4096],
		u_str[4096], 
		n_str[4096],
		e_str[4096],
		pub_ex[5000];
	
	memset (&pgp_info, 0, sizeof(pgp_info));
	
	/* Process the input parameters */
	if( argc < 3 || argc > 5 )
		{
		puts( "Usage: keyex <userID> <outfile> [<pubkey-file>] [<secretkey-file>]" );
		return( EXIT_FAILURE );
		}

	/* Initialise the library */

	status = cryptInit();
	if( cryptStatusError( status ) )
		{
		printf( "cryptInit() failed with error code %d.\n", status );
		return( EXIT_FAILURE );
		}

	/* Check for arguments */
	if(argc > 3)
	{
		strcpy(pubkey, argv[3]);
		if (argc > 4) strcpy(seckey, argv[4]);
		else strcpy(seckey, pubkey);
	} 	
	else
	{
		strcpy(pubkey, PGP_PUBKEY_FILE);
		strcpy(seckey, PGP_PRIVKEY_FILE);
	}
	encryptID = argv[1];
	
	printf("looking for key of %s \n", encryptID);
	printf("pubkey in %s, seckey in %s\n", pubkey, seckey);
	printf("writing results in file <%s>\n\n", argv[2]);
	  
	pub = fopen(pubkey, "r");
	if(pub==NULL) 
	{
		printf("error opening Public Key file\n");
		exit(1);
	} 
	
	out = fopen(argv[2], "w+");
	if(out==NULL) 
	{
		printf("error opening output file\n");
		exit(1);
	} 
	
	gkinfo.keyID = encryptID;
	gkinfo.keyIDsize = strlen(encryptID);
	gkinfo.isKeyID = FALSE;
	gkinfo.isPublicKey = TRUE;

	while( (status = readKey(&pgp_info, pub, &gkinfo)) == -1000);  
	if( cryptStatusError( status ) )
	{
		printf( "Couldn't read Public Key, error code %d.\n",
			status );
		return( EXIT_FAILURE );
	}
	fclose(pub);

	n_to_mpz(e, pgp_info.e, getlen(pgp_info.eLen));
	n_to_mpz(n, pgp_info.n, getlen(pgp_info.nLen));
	
	getstr(n_str, n, pgp_info.nLen);
	getstr(e_str, e, pgp_info.eLen);

	/* assuming one-byte exponent here, should be more intelligent...  */
	
	strcpy(pub_ex, "#0x01");
	strcat(pub_ex, e_str);
	strcat(pub_ex, n_str);
	
	printf("writing public key part to output file...\n");
	fprintf(out, "     pubkey: %s\n", pub_ex);

	/* secret key component */
	
	printf("now trying to read a secret part...\n");
		
	sec = fopen(seckey, "r");
	if(sec==NULL) 
	{
		printf("error opening Secret Key file\n");
		goto end;
	} 

	getPassword( "signature key password", password );
	gkinfo.keyID = encryptID;
	gkinfo.keyIDsize = strlen(encryptID);
	gkinfo.isKeyID = FALSE;
	gkinfo.isPublicKey = FALSE;
	gkinfo.password = password;
	gkinfo.passwordSize = strlen(password);

	while( (status = readKey(&pgp_info, sec, &gkinfo)) == -1000);  

	if( cryptStatusError( status ) )
	{
		printf( "Couldn't read Private Key, error code %d.\n",
			status );
		printf(" No secret key part present, but public key was written!\n");
		goto end;
	}
	fclose(sec);
	
	n_to_mpz(p, pgp_info.p, getlen(pgp_info.pLen));
	n_to_mpz(q, pgp_info.q, getlen(pgp_info.qLen));
	n_to_mpz(d, pgp_info.d, getlen(pgp_info.dLen));

	mpz_init(exp1);
	mpz_sub_ui(t, p, 1);
	mpz_mod(exp1, d, t);		/* exp1 = d mod p-1 */
	mpz_init(exp2);
	mpz_sub_ui(t, q, 1);
	mpz_mod(exp2, d, t);		/* exp2 = d mod q-1 */
	mpz_init(coeff);
	mpz_invert(coeff, q, p);	/* coeff = q^-1 mod p */
	if (mpz_cmp_ui(coeff, 0) < 0)
		mpz_add(coeff, coeff, p);
	passert(mpz_cmp(coeff, p) < 0);

	getstr(d_str, d, pgp_info.dLen);
	getstr(p_str, p, pgp_info.pLen);
	getstr(q_str, q, pgp_info.pLen);
	mpz_get_str(exp1_str, 16, exp1);
	mpz_get_str(exp2_str, 16, exp2);
	mpz_get_str(c_str, 16, coeff);
	
	fprintf(out, "     Modulus: 0x%s\n", n_str);
	fprintf(out, "     PublicExponent: 0x%s\n", e_str);
	fprintf(out, "     PrivateExponent: 0x%s\n", d_str);
	fprintf(out, "     Prime1: 0x%s\n", p_str);
	fprintf(out, "     Prime2: 0x%s\n", q_str);
	fprintf(out, "     Exponent1: 0x%s\n", exp1_str);
	fprintf(out, "     Exponent2: 0x%s\n", exp2_str);
	fprintf(out, "     Coefficient: 0x%s\n", c_str);

	printf("checking key sanity (like Pluto does...)\n");
	RSA_private_key_sanity(n, e, p, q, d, u, exp1, exp2, coeff);
	
end:	fclose(out);
	return( EXIT_SUCCESS );

}

void RSA_private_key_sanity(	mpz_t n,
				mpz_t e, 
				mpz_t p,
				mpz_t q, 
				mpz_t d, 
				mpz_t u_o, 
				mpz_t exp1, 
				mpz_t exp2, 
				mpz_t coeff)
{
    /* note that the *last* error found is reported */
    mpz_t t, u, p1, q1, p1q1, one, de, d2;
    char d2_str[4096];

    mpz_init(t);
    mpz_init(u);
    mpz_init(p1);
    mpz_init(q1);
    mpz_init(p1q1);
    mpz_init(de);
    mpz_init(d2);
    mpz_init(one);
    mpz_add_ui(one, one, 1);

    /* check that n == p * q */
    mpz_mul(u, p, q);
    if (mpz_cmp(u, n) != 0)
	printf("n != p * q\n");

    /* check that e divides neither p-1 nor q-1 */
    mpz_sub_ui(t, p, 1);
    mpz_mod(t, t, e);
    if (mpz_cmp_ui(t, 0) == 0)
	printf("e divides p-1\n");

    mpz_sub_ui(t, q, 1);
    mpz_mod(t, t, e);
    if (mpz_cmp_ui(t, 0) == 0)
	printf("e divides q-1\n");

    /* check that d is e^-1 (mod (p-1) * (q-1)) */
    mpz_sub(u, u, p);  /* u := p*q - p - q +1 */
    mpz_sub(u, u, q);
    mpz_add_ui(u, u, 1);

    /* check that dP is d mod (p-1) */
    mpz_sub_ui(u, p, 1);
    mpz_mod(t, d, u);
    if (mpz_cmp(t, exp1) != 0)
	printf("dP (exp1) is not congruent to d mod (p-1)\n");

    /* check that dQ is d mod (q-1) */
    mpz_sub_ui(u, q, 1);
    mpz_mod(t, d, u);
    if (mpz_cmp(t, exp2) != 0)
	printf("dQ (exp2) is not congruent to d mod (q-1)\n");

    /* check that qInv is (q^-1) mod p */
    mpz_mul(t, coeff, q);
    mpz_mod(t, t, p);
    if (mpz_cmp_ui(t, 1) != 0)
	printf("qInv (coeff) is not conguent ot (q^-1) mod p\n");

    mpz_clear(t);
    mpz_clear(u);
    return;
}
